import os
import time
import uuid
import itertools
import operator
from collections import namedtuple

from sklearn.model_selection import train_test_split

from utils import create_directory_if_not_exist

ArticleSentence = namedtuple('ArticleSentence', 'article_id sentence_id article_title sentence')
SentencePrediction = namedtuple('SentencePrediction', 'article_id sentence_id prediction')

def get_eval_data(dir, remove_whitespace=False):
    x, y = [], []

    dir_filenames = sorted(os.listdir(dir))
    texts_files = filter(is_texts_data_file, dir_filenames)
    labels_files = filter(is_labels_data_file, dir_filenames)

    for texts_file, labels_file in zip(texts_files, labels_files):
        text = get_article_text_data(dir, texts_file)
        label = get_article_label_data(dir, labels_file)
        if (remove_whitespace):
            text_nonempty = []
            label_nonempty = []
            for idx, val in enumerate(text):
                if (val.sentence):
                    text_nonempty.append(text[idx])
                    label_nonempty.append(label[idx])
            text = text_nonempty
            label = label_nonempty
        x.append(text)
        y.append(label)

    return x, y

def get_test_data(dir):
    return [get_article_text_data(dir, texts_file) for texts_file in filter(is_texts_data_file, sorted(os.listdir(dir)))]

def get_article_text_data(dir, texts_file):
    article_id = texts_file[len('article'):-len('.txt')]
    with open(os.path.join(dir, texts_file), 'r', encoding='utf8') as input_file:
        sentences = [line.rstrip('\n') for line in input_file.readlines()]
        title = sentences[0]
        return [ArticleSentence(article_id, sentence_id + 1, title, sentence) for sentence_id, sentence in enumerate(sentences)]

def get_article_label_data(dir, labels_file):
    article_id = labels_file[len('article'):]
    with open(os.path.join(dir, labels_file), 'r', encoding='utf8') as input_file:
        lines = [line.rstrip('\n').split('\t') for line in input_file.readlines()]
        return [label for article_id, sentence_id, label in lines]

def split_data(x, y, ratio=0.80, shuffle=True, stratify_by_labels_ratio=True, random_state=42):
    train_test_split_args = {
        'train_size': ratio,
        'test_size': 1 - ratio,
        'shuffle': shuffle,
        'stratify': y if stratify_by_labels_ratio and shuffle else None,
        'random_state': random_state,
    }

    return train_test_split(x, y, **train_test_split_args)

def persist_predictions(estimator, x, predictions, predictions_dir):
    timestamp = str(int(time.time()))
    model_class_name = estimator.__class__.__name__
    random_uuid = str(uuid.uuid4())

    predictions_folder =  '_'.join([timestamp,  model_class_name, random_uuid])
    predictions_path = os.path.join(predictions_dir, predictions_folder)

    create_directory_if_not_exist(predictions_path)

    print('Writing predictions in:', predictions_path)

    annotated_predictions = []
    for article_sentence, prediction in zip(x, predictions):
        annotated_predictions.append(SentencePrediction(article_sentence.article_id, article_sentence.sentence_id, prediction))

    all_predictions_file_content = []
    for prediction in sorted(annotated_predictions, key= lambda x: x.article_id):
        article_predictions_file_content = '\t'.join([str(prediction.article_id), str(prediction.sentence_id), prediction.prediction])
        all_predictions_file_content.append(article_predictions_file_content)
    all_predictions_file_content = '\n'.join(all_predictions_file_content)

    output_file = os.path.join(predictions_path, 'task2.labels.template')
    with open(output_file, 'w') as predictions_file:
        predictions_file.write(all_predictions_file_content)

def is_texts_data_file(filename):
    return filename.startswith('article') and filename.endswith('.txt')

def is_labels_data_file(filename):
    return filename.startswith('article') and filename.endswith('.task2.labels')
