from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import csv
import os
from bert_modules import modeling as modeling
from bert_modules import optimization as optimization
from bert_modules import tokenization as tokenization
from bert_modules.run_helpers import InputExample, PaddingInputExample, InputFeatures, DataProcessor
from bert_modules.run_helpers import convert_single_example, file_based_convert_examples_to_features, file_based_input_fn_builder
from bert_modules.run_helpers import _truncate_seq_pair, create_model, model_fn_builder, input_fn_builder, convert_examples_to_features
from bert_modules.input_flags import FLAGS

import tensorflow as tf
import numpy as np

from models.base_estimator import BaseEstimator

class Bert(BaseEstimator):
    def __init__(self):
        self.params = {
            'description': 'BERT fine-tuned',
            'bert_params': {
                "data_dir": None, # "The input data dir. Should contain the .tsv files (or other data files) for the task."
                "bert_config_file": 'models/saved/bert/uncased_L-12_H-768_A-12/bert_config.json', # "The config json file corresponding to the pre-trained BERT model. This specifies the model architecture."
                "task_name": 'hack-the-news',  # "The name of the task to train."
                "vocab_file": 'models/saved/bert/uncased_L-12_H-768_A-12/vocab.txt', # "The vocabulary file that the BERT model was trained on."
                "output_dir": 'models/saved/bert_3/', # "The output directory where the model checkpoints will be written."
                "init_checkpoint": 'models/saved/bert_3/model.ckpt-2000', # "Initial checkpoint (usually from a pre-trained BERT model."
                "do_lower_case": True, # "Whether to lower case the input text. Should be True for uncased models and False for cased models."
                "max_seq_length": 64,  # "The maximum total input sequence length after WordPiece tokenization. Sequences longer than this will be truncated, and sequences shorter than this will be padded."
                "do_train": False,  # "Whether to run training."
                "do_eval": False,  # "Whether to run eval on the dev set."
                "do_predict": True,  # "Whether to run the model in inference mode on the test set."
                "train_batch_size": 16,  # "Total batch size for training."
                "eval_batch_size": 8,  # "Total batch size for eval."
                "predict_batch_size": 8,  # "Total batch size for predict."
                "learning_rate": 5e-5,  # "The initial learning rate for Adam."
                "num_train_epochs": 6.0,  # "Total number of training epochs to perform."
                "warmup_proportion": 0.1, # "Proportion of training to perform linear learning rate warmup for. E.g., 0.1 = 10% of training."
                "save_checkpoints_steps": 1000,  # "How often to save the model checkpoint."
                "iterations_per_loop": 1000,  # "How many steps to make in each estimator call."
                "use_tpu": False,  # "Whether to use TPU or GPU/CPU."
                "tpu_name": None,  # "The Cloud TPU to use for training. This should be either the name used when creating the Cloud TPU, or a grpc://ip.address.of.tpu:8470 url."
                "tpu_zone": None, # "[Optional] GCE zone where the Cloud TPU is located in. If not specified, we will attempt to automatically detect the GCE project from metadata."
                "gcp_project": None, # "[Optional] Project name for the Cloud TPU-enabled project. If not specified, we will attempt to automatically detect the GCE project from metadata."
                "master": None,  # "[Optional] TensorFlow master URL."
                "num_tpu_cores": 8,  # "Only used if `use_tpu` is True. Total number of TPU cores to use."
            }
        }

        self.update_tf_flags()
        self.do_train = False
        self.output_dir = 'models/saved/bert_3/'
        self.init_checkpoint = 'models/saved/bert_3/model.ckpt-2000'

    def update_tf_flags(self):
        for key, value in self.params['bert_params'].items(): FLAGS.__setattr__(key, value)

    def fit(self, train_x=None, train_y=None):
        print('BERT !!!!!!!!!!!', train_y)
        # train_x = [x.sentence for x in train_x]

        tf.logging.set_verbosity(tf.logging.INFO)

        processors = {"hack-the-news": HackTheNewsProcessor}

        tokenization.validate_case_matches_checkpoint(FLAGS.do_lower_case, self.init_checkpoint)

        if not self.do_train and not FLAGS.do_eval and not FLAGS.do_predict:
            raise ValueError("At least one of `do_train`, `do_eval` or `do_predict' must be True.")

        bert_config = modeling.BertConfig.from_json_file(FLAGS.bert_config_file)

        if FLAGS.max_seq_length > bert_config.max_position_embeddings:
            raise ValueError(
                "Cannot use sequence length %d because the BERT model "
                "was only trained up to sequence length %d" %
                (FLAGS.max_seq_length, bert_config.max_position_embeddings))

        tf.gfile.MakeDirs(self.output_dir)

        task_name = FLAGS.task_name.lower()

        if task_name not in processors: raise ValueError("Task not found: %s" % (task_name))

        self.processor = processors[task_name]()

        self.label_list = self.processor.get_labels()

        self.tokenizer = tokenization.FullTokenizer(vocab_file=FLAGS.vocab_file, do_lower_case=FLAGS.do_lower_case)

        tpu_cluster_resolver = None
        if FLAGS.use_tpu and FLAGS.tpu_name:
            tpu_cluster_resolver = tf.contrib.cluster_resolver.TPUClusterResolver(
                FLAGS.tpu_name, zone=FLAGS.tpu_zone, project=FLAGS.gcp_project)

        is_per_host = tf.contrib.tpu.InputPipelineConfig.PER_HOST_V2
        run_config = tf.contrib.tpu.RunConfig(
            cluster=tpu_cluster_resolver,
            master=FLAGS.master,
            model_dir=self.output_dir,
            save_checkpoints_steps=FLAGS.save_checkpoints_steps,
            tpu_config=tf.contrib.tpu.TPUConfig(
                iterations_per_loop=FLAGS.iterations_per_loop,
                num_shards=FLAGS.num_tpu_cores,
                per_host_input_for_training=is_per_host))

        train_examples = None
        num_train_steps = None
        num_warmup_steps = None

        if self.do_train:
            train_examples = self.processor.get_train_examples(FLAGS.data_dir, train_x, train_y)
            num_train_steps = int(len(train_examples) / FLAGS.train_batch_size * FLAGS.num_train_epochs)
            num_warmup_steps = int(num_train_steps * FLAGS.warmup_proportion)

        model_fn = model_fn_builder(
            bert_config=bert_config,
            num_labels=len(self.label_list),
            init_checkpoint=self.init_checkpoint,
            learning_rate=FLAGS.learning_rate,
            num_train_steps=num_train_steps,
            num_warmup_steps=num_warmup_steps,
            use_tpu=FLAGS.use_tpu,
            use_one_hot_embeddings=FLAGS.use_tpu)

        # If TPU is not available, this will fall back to normal Estimator on CPU
        # or GPU.
        self.estimator = tf.contrib.tpu.TPUEstimator(
            use_tpu=FLAGS.use_tpu,
            model_fn=model_fn,
            config=run_config,
            train_batch_size=FLAGS.train_batch_size,
            eval_batch_size=FLAGS.eval_batch_size,
            predict_batch_size=FLAGS.predict_batch_size)

        if self.do_train:
            train_file = os.path.join(self.output_dir, "train.tf_record")
            file_based_convert_examples_to_features(train_examples, self.label_list, FLAGS.max_seq_length, self.tokenizer, train_file)
            tf.logging.info("***** Running training *****")
            tf.logging.info("  Num examples = %d", len(train_examples))
            tf.logging.info("  Batch size = %d", FLAGS.train_batch_size)
            tf.logging.info("  Num steps = %d", num_train_steps)
            train_input_fn = file_based_input_fn_builder(
                input_file=train_file,
                seq_length=FLAGS.max_seq_length,
                is_training=True,
                drop_remainder=True)
            self.estimator.train(input_fn=train_input_fn, max_steps=num_train_steps)

    def predict(self, test_x, return_probabilities = False):
        test_x = [x.sentence for x in test_x]
        if FLAGS.do_predict:
            predict_examples = self.processor.get_test_examples(FLAGS.data_dir, test_x)
            num_actual_predict_examples = len(predict_examples)
            if FLAGS.use_tpu:
                # TPU requires a fixed batch size for all batches, therefore the number
                # of examples must be a multiple of the batch size, or else examples
                # will get dropped. So we pad with fake examples which are ignored
                # later on.
                while len(predict_examples) % FLAGS.predict_batch_size != 0:
                    predict_examples.append(PaddingInputExample())

            predict_file = os.path.join(self.output_dir, "predict.tf_record")
            file_based_convert_examples_to_features(predict_examples, self.label_list, FLAGS.max_seq_length, self.tokenizer, predict_file)

            tf.logging.info("***** Running prediction*****")
            tf.logging.info("  Num examples = %d (%d actual, %d padding)",
                            len(predict_examples), num_actual_predict_examples,
                            len(predict_examples) - num_actual_predict_examples)
            tf.logging.info("  Batch size = %d", FLAGS.predict_batch_size)

            predict_drop_remainder = True if FLAGS.use_tpu else False
            predict_input_fn = file_based_input_fn_builder(
                input_file=predict_file,
                seq_length=FLAGS.max_seq_length,
                is_training=False,
                drop_remainder=predict_drop_remainder)

            result = list(self.estimator.predict(input_fn=predict_input_fn))

            prediction_probs = [prediction['probabilities'].tolist() for prediction in result]
            predictions = [self.label_list[probabilities.index(max(probabilities))] for probabilities in prediction_probs]

            output_predict_file = os.path.join(self.output_dir, "test_results.tsv")
            with tf.gfile.GFile(output_predict_file, "w") as writer:
                num_written_lines = 0
                tf.logging.info("***** Predict results *****")
                for (i, prediction) in enumerate(result):
                    probabilities = prediction["probabilities"]
                    if i >= num_actual_predict_examples: break
                    output_line = "\t".join(
                        str(class_probability)
                        for class_probability in probabilities) + "\n"
                    writer.write(output_line)
                    num_written_lines += 1
            assert num_written_lines == num_actual_predict_examples

            if return_probabilities: return np.array(prediction_probs)

            return np.array(predictions)

class HackTheNewsProcessor(DataProcessor):
    def get_train_examples(self, data_dir, x, y):
        examples = []
        for i, (text, label) in enumerate(zip(x, y)):
            guid = "train-%d" % (i)
            text_a = tokenization.convert_to_unicode(text)
            text_b = None
            label = tokenization.convert_to_unicode(label)

            examples.append(InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
        return examples

    def get_test_examples(self, data_dir, x):
        examples = []
        for i, text in enumerate(x):
            guid = "test-%d" % (i)
            text_a = tokenization.convert_to_unicode(text)
            text_b = None
            label = "non-propaganda"

            examples.append(InputExample(guid=guid, text_a=text_a, text_b=text_b, label=label))
        return examples

    def get_labels(self):
        return ["non-propaganda", "propaganda"]
