from sklearn.linear_model import LogisticRegression
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.svm import SVC
from sklearn.neural_network import MLPClassifier
from sklearn.preprocessing import StandardScaler, MinMaxScaler

import numpy as np

from models.base_estimator import BaseEstimator
from utils import flatten, get_n_jobs

from transformers import lexical_per_sentence
from transformers.loaded_language import loaded_language_phrase_counts, loaded_language_phrases_count
from transformers import readability
from transformers import compute_sentiment
from transformers import polarity
from transformers import subjectivity
from transformers import EmotionsExtractor
from transformers import has_confusing_words
import itertools
import scipy

def read_loaded_language_phrases():
    with open('data/external/loaded_language_phrases.txt', 'r', encoding='utf-8') as file:
        phrases = [line.rstrip('\n') for line in file.readlines()]
    return phrases


class LogRegAll(BaseEstimator):
    def __init__(self):
        self.params = {
            'description': 'MLP with many types of feaures',
            'logreg_clf_params': {
                'C': 0.6,
                'penalty': 'l2',
                'class_weight': 'balanced',
                'solver': 'lbfgs',
                'max_iter': 10000,
            },
            'ab_params': {
                'n_estimators': 300,
                'random_state': 42,
            },
            'svm_params': {
                'C': 0.7,
                'class_weight': 'balanced',
                'kernel': 'linear',
                'tol': 0.001,
                'verbose': 1,
                'max_iter': -1,
                'random_state': 42,
                'probability': True
            },
            'mlp_params': {
                'hidden_layer_sizes': (200, 300, 200),
                'max_iter': 10000,
                'verbose': 1,
                'tol': 0.001,
                'alpha': 0.0001,
                'solver': 'adam'
            },
        }

        self.phrases = read_loaded_language_phrases()

    def fit(self, train_x, train_y):
        train_x_sent = train_x
        train_x = [x.sentence for x in train_x]

        self.emotions_extractor = EmotionsExtractor()
        self.vectorizer = TfidfVectorizer(sublinear_tf=True)
        self.scaler = StandardScaler(with_mean=True, with_std=True)

        emotions_features = self.emotions_extractor.load_features(train_x_sent)
        # tfidf_features = self.vectorizer.fit_transform(train_x)
        lexical_features = lexical_per_sentence(train_x)
        loaded_lang_features = [[loaded_language_phrases_count(sentence, self.phrases)] for sentence in train_x]
        readability_features = readability(train_x)
        sentiment_features = compute_sentiment(train_x)
        subjectivity_features = [[subjectivity(x)] for x in train_x]
        polarity_features = [[polarity(x)] for x in train_x]
        confusing_words_features = has_confusing_words(train_x)

        sentences_features = scipy.sparse.hstack([
            emotions_features,
            # tfidf_features,
            lexical_features,
            loaded_lang_features,
            readability_features,
            sentiment_features,
            subjectivity_features,
            polarity_features,
            confusing_words_features
        ])

        sentences_features = self.scaler.fit_transform(sentences_features)
        self.model = LogisticRegression(**self.params['logreg_clf_params'])
        self.model.fit(sentences_features, train_y)

    def predict(self, test_x):
        test_x_sent = test_x
        test_x = [x.sentence for x in test_x]

        emotions_features = self.emotions_extractor.load_features(test_x_sent)
        # tfidf_features = self.vectorizer.transform(test_x)
        lexical_features = lexical_per_sentence(test_x)
        loaded_lang_features = [[loaded_language_phrases_count(sentence, self.phrases)] for sentence in test_x]
        readability_features = readability(test_x)
        sentiment_features = compute_sentiment(test_x)
        subjectivity_features = subjectivity(test_x)
        polarity_features = polarity(test_x)
        confusing_words_features = has_confusing_words(train_x)

        sentences_features = scipy.sparse.hstack([
            emotions_features,
            # tfidf_features,
            lexical_features,
            loaded_lang_features,
            readability_features,
            sentiment_features,
            subjectivity_features,
            polarity_features,
            confusing_words_features
        ])
        sentences_features = self.scaler.transform(sentences_features)
        return self.model.predict(sentences_features)
