import numpy as np
import scipy
import gensim
from sklearn.linear_model import LogisticRegression, SGDClassifier
from sklearn.svm import SVC
from sklearn.ensemble import RandomForestClassifier, AdaBoostClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.tree import DecisionTreeClassifier
from sklearn.metrics import f1_score
from sklearn.feature_extraction.text import TfidfVectorizer
from nltk.corpus import stopwords
from nltk.tokenize import RegexpTokenizer

from models.base_estimator import BaseEstimator
from transformers import readability
from transformers import compute_sentiment, lexical_per_sentence
from transformers import EmotionsExtractor, merge_pos_counts, subjectivity, polarity
from utils import flatten, get_n_jobs
from transformers import loaded_language_phrases_count, has_confusing_words
from transformers.proper_nouns import singular_proper_noun_max_counts

from models import Bert, BertUntuned

TASK_2_POSITIVE_LABEL = 'propaganda'

WORD2VEC_EMBEDDINGS_PATH = 'data/external/embeddings/GoogleNews-vectors-negative300.bin'

def read_loaded_language_phrases():
    with open('data/external/loaded_language_phrases.txt', 'r', encoding='utf-8') as file:
        phrases = [line.rstrip('\n') for line in file.readlines()]
    return phrases

class Stacking(BaseEstimator):
    def __init__(self):
        self.params = {
            'model_description': 'Meta Stacking',
            'meta_learner_train_ratio': 0.75,
            'meta_learner_prob': True,
            'meta_params': {
                'penalty': 'l2',
                'C': 1,
                'class_weight': 'balanced',
                'solver': 'lbfgs',
                'max_iter': 3000,
            },
            'log_reg_params': {
                'penalty': 'l2',
                'C': 0.8,
                'class_weight': 'balanced',
                'solver': 'lbfgs',
                'max_iter': 3000,
                'n_jobs': 8,
            },
            'emotions_log_reg_params': {
                'penalty': 'l2',
                'dual': False,
                'tol': 0.0001,
                'C': 0.8,
                'fit_intercept': True,
                'intercept_scaling': 1,
                'class_weight': 'balanced',
                'random_state': None,
                'solver': 'liblinear',
                'max_iter': 3000,
                'multi_class': 'ovr',
                'verbose': 0,
                'warm_start': False,
                'n_jobs': 8,
            },
            'scaler_params': {
                'with_mean': True,
                'with_std': True
            },
            'readability_params': {
                'indices': [
                    'flesch_reading_ease',
                    'smog_index',
                    'flesch_kincaid_grade',
                    'coleman_liau_index',
                    'automated_readability_index',
                    'dale_chall_readability_score',
                    'difficult_words',
                    'linsear_write_formula',
                    'gunning_fog'
                 ]
            },
        }

        self.phrases = read_loaded_language_phrases()
        self.emotions_extractor = EmotionsExtractor()
        self.vectorizer = TfidfVectorizer(sublinear_tf=True)
        self.tokenizer = RegexpTokenizer(r'\w+')
        self.stop_words = set(stopwords.words('english'))
        self.word2vec_model = gensim.models.KeyedVectors.load_word2vec_format(WORD2VEC_EMBEDDINGS_PATH, binary=True)

    def fit(self, train_x, train_y):
        map_sentences = lambda x: [article.sentence for article in x]

        loaded_lang_func = lambda x: np.array(
            [[loaded_language_phrases_count(article.sentence, self.phrases)] for article in x]
        )

        proper_nouns_func = lambda x: singular_proper_noun_max_counts(map_sentences(x))

        confusing_words_func = lambda x: has_confusing_words(map_sentences(x))

        lexical_features_func = lambda x: lexical_per_sentence(map_sentences(x))

        readability_func = lambda x: readability(map_sentences(x), **self.params['readability_params'])

        sentiment_func = lambda x: compute_sentiment(map_sentences(x))

        polarity_scalar = StandardScaler(**self.params['scaler_params'])
        subjectivity_func = lambda x: np.concatenate((
            list(subjectivity(map_sentences(x))),
            polarity_scalar.transform(list(polarity(map_sentences(x)))),
        ), axis = 1)

        sent_len_func = lambda x: np.array([len(article.sentence) for article in x]).reshape(-1, 1)

        emotions_func = lambda x: self.emotions_extractor.load_features(x)

        tfidf_func = lambda x: self.vectorizer.transform(map_sentences(x))

        word2vec_func = lambda x: [self.avg(self.clean_stop_words(article.sentence)) for article in x]

        self.stack = [
            (loaded_lang_func, StandardScaler(**self.params['scaler_params']), LogisticRegression(**self.params['log_reg_params'])),
            (proper_nouns_func, StandardScaler(**self.params['scaler_params']), LogisticRegression(**self.params['log_reg_params'])),
            (confusing_words_func, None, LogisticRegression(**self.params['log_reg_params'])),
            (lexical_features_func, None, LogisticRegression(**self.params['log_reg_params'])),
            (readability_func, None, LogisticRegression(**self.params['log_reg_params'])),
            (sentiment_func, None, LogisticRegression(**self.params['log_reg_params'])),
            (subjectivity_func, None, LogisticRegression(**self.params['log_reg_params'])),
            (sent_len_func, StandardScaler(**self.params['scaler_params']), LogisticRegression(**self.params['log_reg_params'])),
            (emotions_func, None, LogisticRegression(**self.params['emotions_log_reg_params'])),
            (tfidf_func, None, LogisticRegression(**self.params['log_reg_params'])),
            (word2vec_func, None, LogisticRegression(**self.params['log_reg_params'])),
        ]

        train_x_zero, train_y_zero, train_x_meta, train_y_meta = self.split_data(train_x, train_y)
        polarity_scalar.fit(list(polarity(map_sentences(train_x_zero))))
        self.vectorizer.fit(map_sentences(train_x_zero))
        predictions_zero = self.stack_fit_predict(train_x_zero, train_y_zero, train_x_meta, train_y_meta)

        self.meta_learner = LogisticRegression(**self.params['meta_params'])
        self.meta_learner.fit(self.convert_to_meta_input(predictions_zero), train_y_meta)

        self.stack_fit_predict(train_x, train_y)

    def predict(self, test_x):
        predictions = []

        print('Predicting - Bert Test')
        bert_test = Bert()
        bert_test.fit()
        bert_probs = bert_test.predict(test_x, return_probabilities=True).tolist()
        predictions.append(bert_probs)

        for transformer, scaler, predictor in self.stack:
            data_transformed = transformer(test_x)
            data_transformed = scaler.transform(data_transformed) if scaler else data_transformed

            prediction = predictor.predict_proba(data_transformed).tolist()
            predictions.append(prediction)

        prediction_converted = self.convert_to_meta_input(predictions)
        return self.meta_learner.predict(prediction_converted)

    def stack_fit_predict(self, train_x, train_y, test_x = None, test_y = None):
        predictions = []

        if test_x:
            with open('data/bert_meta_results.tsv', 'r') as input_file:
                bert_probs = [[float(x) for x in line.rstrip('\n').split()] for line in input_file.readlines()]

            # bert_meta = BertUntuned()
            # print('Fitting - Bert Meta')
            # bert_meta.fit(train_x, train_y)
            # print('Predicting - Bert Meta')
            # bert_probs = bert_meta.predict(test_x, return_probabilities=True).tolist()
            predictions.append(bert_probs)

        for transformer, scaler, predictor in self.stack:
            data_transformed_zero = transformer(train_x)
            data_transformed_zero = scaler.fit_transform(data_transformed_zero) if scaler else data_transformed_zero

            if (test_x):
                data_transformed_meta = transformer(test_x)
                data_transformed_meta = scaler.transform(data_transformed_meta) if scaler else data_transformed_meta

            predictor.fit(data_transformed_zero, train_y)

            if (test_x):
                prediction = predictor.predict_proba(data_transformed_meta).tolist()
                predictions.append(prediction)

        if (test_x):
            return predictions

    def convert_to_meta_input(self, predictions):
        # s[0] -> non-propaganda probability
        return list(map(
            lambda sample: list(map(lambda s: s[0] if self.params['meta_learner_prob'] else s[0] < s[1], sample)),
            list(zip(*predictions))
        ))

    def split_data(self, train_x, train_y):
        split_pos = round(len(train_x) * self.params['meta_learner_train_ratio'])

        train_x_zero = train_x[:split_pos]
        train_x_meta = train_x[split_pos:]

        train_y_zero = train_y[:split_pos]
        train_y_meta = train_y[split_pos:]

        return train_x_zero, train_y_zero, train_x_meta, train_y_meta

    def avg(self, tokenized_sentence):
        a = [self.word2vec_model.wv[word] for word in tokenized_sentence if word in self.word2vec_model.wv]
        if (len(a) == 0):
            return [0 for x in range(0, 300)]
        else:
            return np.average(a, 0)

    def clean_stop_words(self, text):
        word_tokens = self.tokenizer.tokenize(text)
        return [w for w in word_tokens if not w in self.stop_words]
