import numpy as np
import transformers
from sklearn.svm import SVC
from models.base_estimator import BaseEstimator

class SVMSubjectivity(BaseEstimator):
    def __init__(self):
        self.params = {
            'description': 'SVM with subjectivity and polarity',
            'svc_clf_params': {
                'C': 1.0,
                'cache_size': 200,
                'class_weight': None,
                'coef0': 0.0,
                'decision_function_shape': 'ovr',
                'degree': 3,
                'gamma': 'auto',
                'kernel': 'linear',
                'max_iter': -1,
                'probability': False,
                'random_state': 42,
                'shrinking': True,
                'tol': 0.001,
                'verbose': False,
            },
            'readability_indices': ['flesch_reading_ease', 'smog_index', 'difficult_words']
        }

    def fit(self, train_x, train_y):
        train_x = [x.sentence for x in train_x]

        features = self.transform_features(train_x)

        self.model = SVC(**self.params['svc_clf_params'])
        self.model.fit(features, train_y)

    def predict(self, test_x):
        test_x = [x.sentence for x in test_x]

        features = self.transform_features(test_x)

        return self.model.predict(features)

    def transform_features(self, data_x):
        features = np.concatenate((
            transformers.subjectivity(data_x),
            transformers.polarity(data_x),
            transformers.count_adjectives(data_x),
            transformers.count_adverbs(data_x),
            transformers.readability(data_x, indices=self.params['readability_indices'])
        ), axis = 1)
        return features

    def get_grid_params(self):
        return {
            'C': (0.5, 1.0),
            'kernel': ('linear', 'poly', 'rbf', 'sigmoid', 'precomputed')
        }
