import traceback
import pandas as pd
import os
from watson_developer_cloud import NaturalLanguageUnderstandingV1
from watson_developer_cloud.natural_language_understanding_v1 import Features, EmotionOptions
import time
import json

dirname = os.path.dirname(__file__)
config_path = os.path.join(dirname, '../config.json')
with open(config_path) as fp:
    config = json.load(fp)

INDICO_EMOTION_FIELDS = ['fear', 'anger', 'surprise', 'sadness', 'joy']
IBM_EMOTION_FIELDS = ['sadness', 'joy', 'fear', 'disgust', 'anger']

class EmotionsExtractor:
    def __init__(self):
        self.load_cache()

    def save_to_csv(self, article_sentences, emotions):
        data = {
            field: [emotion[field] for emotion in emotions] for field in emotions[0]
        }
        data['sentence'] = [a.sentence for a in article_sentences]
        data['article_id'] = [a.article_id for a in article_sentences]
        data['sentence_id'] = [a.sentence_id for a in article_sentences]
        df = pd.DataFrame(data).set_index(['article_id', 'sentence_id'])
        timestamp = str(int(time.time()))
        df.to_csv(os.path.join(dirname,'./emotions_cache/data-{}.csv'.format(timestamp)))

    def load_cache(self):
        self.cache = pd.read_csv(os.path.join(dirname, './emotions_cache/all_emotions.csv')).set_index(['article_id', 'sentence_id'])
        self.cache = self.cache.sort_index()

    def save_features(self, article_sentences):
        emotions = self.__get_features_from_service([a.sentence for a in article_sentences])
        self.save_to_csv(article_sentences, emotions)

    def __get_features_from_service(self, sentences):
        return self.__get_ibm_features(sentences)

    def __get_ibm_features(self, sentences):
        return [self.__get_ibm_single_sentence(sentence, idx) for idx, sentence in enumerate(sentences)]

    def __get_ibm_single_sentence(self, sentence, idx):
        print('Handling sentence ', idx, ': ', sentence)
        ibm_api_key = config['ibm_api_key']
        # The text to analyze
        natural_language_understanding = NaturalLanguageUnderstandingV1(
            version='2018-11-16',
            iam_apikey=ibm_api_key,
            url='https://gateway-lon.watsonplatform.net/natural-language-understanding/api'
        )

        response = natural_language_understanding.analyze(
            language='en',
            text=sentence,
            features=Features(emotion=EmotionOptions(document=True))).get_result()
        return response['emotion']['document']['emotion']

    def __get_indico_features(self, sentences):
        import indicoio as indico
        return indico.emotion(sentences)

    def load_features(self, sentences):
        return self.cache.loc[[(int(s.article_id), s.sentence_id) for s in sentences]].values
