import os
import errno
import json
from itertools import zip_longest, chain
from collections import OrderedDict

import numpy as np

from sklearn.metrics import f1_score, precision_score, recall_score, accuracy_score

TASK_2_POSITIVE_LABEL = 'propaganda'

def empty_sentences_mask(x, y=None):
    new_x, new_y = [], []
    mask_index = []
    mask_sent = []
    for index, sentence in enumerate(x):
        alphabetical_char_count = sum(c.isalpha() for c in sentence.sentence)
        if alphabetical_char_count == 0:
            mask_index.append(index)
            mask_sent.append(sentence)
        else:
            new_x.append(sentence)
            if y: new_y.append(y[index])

    return new_x, new_y, mask_index, mask_sent

def insert_empty_sentences_mask(predictions, mask_index, x = None, mask_sentence = None):
    for index, prediction_index in enumerate(mask_index):
        predictions.insert(prediction_index, 'non-propaganda')
        if x: x.insert(prediction_index, mask_sentence[index])

    return predictions, x

def compute_score(submission_labels, gold_labels):
    accuracy = accuracy_score(gold_labels, submission_labels)
    precision_pos = precision_score(gold_labels, submission_labels, pos_label=TASK_2_POSITIVE_LABEL)
    recall_pos = recall_score(gold_labels, submission_labels, pos_label=TASK_2_POSITIVE_LABEL)
    f1_pos = f1_score(gold_labels, submission_labels, pos_label=TASK_2_POSITIVE_LABEL)

    return accuracy, precision_pos, recall_pos, f1_pos

def aggregate_results(clf_params, cv=None, val=None, test=None):
    if cv:
        cv = {
            'train_score': {
                'mean': round(np.mean(cv['train_score']), 4),
                'std': round(np.std(cv['train_score']), 2)
            },
            'test_score': {
                'mean': round(np.mean(cv['test_score']), 4),
                'std': round(np.std(cv['test_score']), 2),
                'all': round_np_scores(cv['test_score'], 4)
            },
            'fit_time': humanize_time(max(cv['fit_time'])),
            'score_time': humanize_time(max(cv['score_time'])),
            'size': cv['size'],
            'folds': cv['folds'],
        }

    if val:
        val = {
            'accuracy': round(val['accuracy'], 4),
            'precision_pos': round(val['precision_pos'], 4),
            'recall_pos': round(val['recall_pos'], 4),
            'f1_pos': round(val['f1_pos'], 4),
            'time': humanize_time(val['time']),
            'train_size': val['train_size'],
            'validation_size': val['validation_size'],
        }

    results = {
        'estimator': clf_params,
        'cross_validation': cv,
        'validation': val,
        'test': test,
    }

    results_json = json.dumps(OrderedDict(**results), indent=4)
    print(results_json)

    return results_json

def write_results_to_file(results):
    results_path = config_local().get('results_file', None)

    if not results_path: print('No results_file specified in the local configuration!'); return

    create_directory_if_not_exist(os.path.dirname(results_path))
    file_prepend(results_path, results, '\n' + '='*100 + '\n')

def create_directory_if_not_exist(directory):
    try:
        os.makedirs(directory)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

def file_prepend(file_path, content, separator = ''):
    try:
        with open(file_path, 'r+') as output:
            old_content = output.read()
            if not old_content: separator = ''
            new_content = content + separator + old_content
            output.seek(0, 0)
            output.write(new_content)
    except FileNotFoundError:
        with open(file_path, 'w') as output:
            output.write(content)

def config_local():
    with open('config.json', 'r') as config_json:
        return json.load(config_json)

def round_np_scores(np_array, p=None):
    return [round(x, p) for x in np_array.tolist()]

def humanize_time(secs):
    mins, secs = divmod(secs, 60)
    hours, mins = divmod(mins, 60)

    return '%02d:%02d:%02d' % (hours, mins, secs)

def print_progress_bar(iteration, total, description='', decimals=1, bar_length=100, fill='█'):
    percent = ('{0:.' + str(decimals) + 'f}').format(100 * (iteration / float(total)))
    filled_length = int(bar_length * iteration // total)
    bar = fill * filled_length + '-' * (bar_length - filled_length)

    print('\r |%s| %s%% | %s' % (bar, percent, description), end='\r')

    if iteration == total: print()

def chunker(iterable, n, fillvalue=None):
    args = [iter(iterable)] * n

    return list(zip_longest(*args, fillvalue=fillvalue))

def get_n_jobs():
    with open('config.json', 'r') as rc:
        return json.load(rc).get('n_jobs', 1)

def update_dict(params, keys, value):
        if len(keys) == 1:
            params[keys[0]] = value
        else:
            update_dict(params[keys[0]], keys[1:], value)

def flatten(list):
    return [item for sublist in list for item in sublist]

def merge_dataset_nested_lists(a, b):
    return [list(chain.from_iterable(x)) for x in zip(a, b)]
